@@ -244,9 +244,9 @@ gsf_xml_parser_context_full (GsfInput *input, xmlSAXHandlerPtr sax, gpointer use
  * Create a libxml2 pull style parser context wrapper around a gsf input.
  * This signature will probably change to supply a SAX structure.
  *
- * NOTE : adds a reference to @input
- * NOTE : a simple wrapper around a cleaner implementation that will fold in
- *	 when we add other api changes.  Its not worth bumping just for this
+ * <note>This adds a reference to @input.</note>
+ * <note>A simple wrapper around a cleaner implementation that will fold in
+ * when we add other api changes.  Its not worth bumping just for this.</note>
  *
  * Returns : A parser context or NULL
  **/
@@ -258,11 +258,11 @@ gsf_xml_parser_context (GsfInput *input)
 
 /**
  * gsf_xml_output_buffer_new :
- * @output :
- * @encoding : optionally NULL.
+ * @output:
+ * @encoding: optionally NULL.
  *
- * NOTE : adds a reference to @output
- * NOTE : This is _not_ releated to GsfXMLOut
+ * <note>This adds a reference to @output.</note>
+ * <note>This is <emphasis>not</emphasis> releated to #GsfXMLOut.</note>
  */
 static xmlOutputBufferPtr
 gsf_xml_output_buffer_new (GsfOutput *output,
@@ -322,6 +322,7 @@ typedef struct {
 	GsfXMLInNode pub;
 	/* internal state */
 	GSList *groups;
+	GSList *extensions;
 } GsfXMLInNodeInternal;
 struct _GsfXMLInDoc {
 	GsfXMLInNodeInternal const *root_node;
@@ -333,14 +334,14 @@ struct _GsfXMLInDoc {
 typedef struct {
 	GsfXMLIn	pub;
 
-	GsfXMLInNS   const *default_ns;	/* optionally NULL */
-	GSList	 	   *ns_stack;
-	GHashTable	   *ns_prefixes;
-	GPtrArray	   *ns_by_id;
-	gint	  	    unknown_depth; /* handle recursive unknown tags */
-	gboolean	    from_unknown_handler;
+	GsfXMLInNS const *default_ns;	   /* optionally NULL */
+	GSList	 	 *ns_stack;
+	GHashTable	 *ns_prefixes;
+	GPtrArray	 *ns_by_id;
+	gint	  	  unknown_depth; /* handle recursive unknown tags */
+	gboolean	  from_unknown_handler;
 
-	GSList	 	   *extension_stack;
+	GSList	 	 *extension_stack; /* stack of GsfXMLInExtension */
 } GsfXMLInInternal;
 typedef struct {
 	char    *tag;
@@ -352,9 +353,9 @@ typedef struct {
 	GSList *elem;
 } GsfXMLInNodeGroup;
 typedef struct {
-	GsfXMLInDoc const *old_doc;
-	gpointer	   old_state;
-	GsfXMLInExtDtor    dtor;
+	GsfXMLInExtDtor	   dtor;
+	gpointer	   state;
+	GsfXMLInDoc const *doc;
 	gboolean	   from_unknown;
 } GsfXMLInExtension;
 
@@ -376,20 +377,32 @@ push_child (GsfXMLInInternal *state, GsfXMLInNode const *node, GsfXMLInNS const
 		(gpointer)state->pub.node);
 	state->ns_stack		= g_slist_prepend (state->ns_stack,
 		(gpointer)state->default_ns);
-	state->extension_stack	= g_slist_prepend (state->extension_stack, ext);
 	state->pub.node = node;
 	state->default_ns = default_ns;
-	if (node->start != NULL)
+
+	state->extension_stack	= g_slist_prepend (state->extension_stack, ext);
+	if (NULL != ext) {
+		GsfXMLInDoc const *old_doc = state->pub.doc;
+		state->pub.doc = ext->doc;
+		ext->doc = old_doc;
+		if (NULL != ext->state) {
+			gpointer old_state = state->pub.user_state;
+			state->pub.user_state = ext->state;
+			ext->state = old_state;
+		}
+	}
+	if (NULL != node->start)
 		node->start (&state->pub, attrs);
 }
 
 static gboolean
 lookup_child (GsfXMLInInternal *state, GsfXMLInNS const *default_ns,
-	      GsfXMLInNSInstance *inst,
-	      GSList *groups, xmlChar const *name, xmlChar const **attrs)
+	      GSList *groups, xmlChar const *name,
+	      xmlChar const **attrs, GsfXMLInExtension *ext)
 {
 	GsfXMLInNodeGroup  *group;
 	GsfXMLInNode	   *node;
+	GsfXMLInNSInstance *inst;
 	GSList *elem, *ptr;
 	char const *tmp;
 
@@ -413,7 +426,7 @@ lookup_child (GsfXMLInInternal *state, GsfXMLInNS const *default_ns,
 		for (elem = group->elem ; elem != NULL ; elem = elem->next) {
 			node = elem->data;
 			if (node->name == NULL || !strcmp (tmp, node->name)) {
-				push_child (state, node, default_ns, attrs, NULL);
+				push_child (state, node, default_ns, attrs, ext);
 				return TRUE;
 			}
 		}
@@ -478,14 +491,22 @@ gsf_xml_in_start_element (GsfXMLInInternal *state, xmlChar const *name, xmlChar
 	}
 
 	node = (GsfXMLInNodeInternal const *) state->pub.node;
-	if (lookup_child (state, default_ns, inst, node->groups, name, attrs))
+	if (lookup_child (state, default_ns, node->groups, name, attrs, NULL))
 		return;
 
 	/* useful for <Data><b><i><u></u></i></b></Data> where all of the markup can nest */
 	ptr = state->pub.node_stack;
 	for (; ptr != NULL && node->pub.share_children_with_parent; ptr = ptr->next) {
 		node = ptr->data;
-		if (lookup_child (state, default_ns, inst, node->groups, name, attrs))
+		if (lookup_child (state, default_ns, node->groups, name, attrs, NULL))
+			return;
+	}
+
+	/* Check for extensions */
+	for (ptr = node->extensions; ptr != NULL ; ptr = ptr->next) {
+		GsfXMLInExtension *ext = ptr->data;
+		if (lookup_child (state, default_ns,
+			ext->doc->root_node->groups, name, attrs, ext))
 			return;
 	}
 
@@ -515,46 +536,63 @@ gsf_xml_in_start_element (GsfXMLInInternal *state, xmlChar const *name, xmlChar
 	state->pub.node_stack = g_slist_reverse (state->pub.node_stack);
 }
 
+static void
+gsf_xml_in_ext_free (GsfXMLInInternal *state, GsfXMLInExtension *ext)
+{
+	if (ext->dtor)
+		(ext->dtor) (&state->pub, ext->state);
+	g_free (ext);
+}
+
 static void
 gsf_xml_in_end_element (GsfXMLInInternal *state,
 			G_GNUC_UNUSED xmlChar const *name)
 {
-	GsfXMLInExtension *ext;
-	gboolean was_overlay;
+	GsfXMLInNodeInternal	*node;
+	GsfXMLInExtension	*ext;
+	GSList *ptr;
 
 	if (state->unknown_depth > 0) {
 		state->unknown_depth--;
 		return;
 	}
 
-	do {
-		g_return_if_fail (state->pub.node_stack != NULL);
-		g_return_if_fail (state->ns_stack != NULL);
-		g_return_if_fail (state->extension_stack != NULL);
+	g_return_if_fail (state->pub.node	!= NULL);
+	g_return_if_fail (state->pub.node_stack != NULL);
+	g_return_if_fail (state->ns_stack != NULL);
 
-		if (state->pub.node->end)
-			state->pub.node->end (&state->pub, NULL);
-		/* pop the state stack */
-		ext = state->extension_stack->data;
-		state->extension_stack	= g_slist_remove (state->extension_stack, ext);
-		state->pub.node		= state->pub.node_stack->data;
-		state->pub.node_stack	= g_slist_remove (state->pub.node_stack, state->pub.node);
-		state->default_ns	= state->ns_stack->data;
-		state->ns_stack		= g_slist_remove (state->ns_stack, state->default_ns);
-
-		if (NULL != ext) {
-			was_overlay = !ext->from_unknown;
-			if (ext->dtor)
-				(ext->dtor) (&state->pub, ext->old_state);
-			state->pub.user_state = ext->old_state;
-			state->pub.doc = ext->old_doc;
-			g_free (ext);
-		} else
-			was_overlay = FALSE;
-	} while (was_overlay);
-
-	if (state->pub.node->has_content == GSF_XML_CONTENT)
+	node = (GsfXMLInNodeInternal *) state->pub.node;
+	if (node->pub.end)
+		node->pub.end (&state->pub, NULL);
+	if (node->pub.has_content == GSF_XML_CONTENT)
 		g_string_truncate (state->pub.content, 0);
+
+	/* Free any potential extensions associated with the current node */
+	for (ptr = node->extensions; ptr != NULL ; ptr = ptr->next)
+		gsf_xml_in_ext_free (state, ptr->data);
+	g_slist_free (node->extensions);
+	node->extensions = NULL;
+
+	/* pop the state stack */
+	ext = state->extension_stack->data;
+	state->extension_stack	= g_slist_remove (state->extension_stack, ext);
+	state->pub.node		= state->pub.node_stack->data;
+	state->pub.node_stack	= g_slist_remove (state->pub.node_stack, state->pub.node);
+	state->default_ns	= state->ns_stack->data;
+	state->ns_stack		= g_slist_remove (state->ns_stack, state->default_ns);
+
+	if (NULL != ext) {
+		GsfXMLInDoc const *ext_doc = state->pub.doc;
+		state->pub.doc = ext->doc;
+		ext->doc = ext_doc;
+		if (NULL != ext->state) {
+			gpointer ext_state = state->pub.user_state;
+			state->pub.user_state = ext->state;
+			ext->state = ext_state;
+		}
+		if (ext->from_unknown)
+			gsf_xml_in_ext_free (state, ext);
+	}
 }
 
 static void
@@ -582,6 +620,7 @@ gsf_xml_in_start_document (GsfXMLInInternal *state)
 	state->ns_by_id		= g_ptr_array_new ();
 	state->ns_prefixes	= g_hash_table_new_full (
 		g_str_hash, g_str_equal, NULL, g_free);
+	state->from_unknown_handler = FALSE;
 }
 
 static void
@@ -675,6 +714,10 @@ gsf_xml_in_node_internal_free (GsfXMLInNodeInternal *node)
 	GSList *ptr;
 	GsfXMLInNodeGroup *group;
 
+	if (NULL != node->extensions) {
+		g_warning ("leaking extensions");
+	}
+
 	for (ptr = node->groups; ptr != NULL ; ptr = ptr->next) {
 		group = ptr->data;
 		g_slist_free (group->elem);
@@ -745,7 +788,7 @@ gsf_xml_in_doc_new (GsfXMLInNode const *nodes, GsfXMLInNS const *ns)
 
 	for (e_node = nodes; e_node->id != NULL ; e_node++ ) {
 		node = g_hash_table_lookup (doc->symbols, e_node->id);
-		if (tmp != NULL) {
+		if (node != NULL) {
 			/* if its empty then this is just a recusion */
 			if (e_node->start != NULL || e_node->end != NULL ||
 			    e_node->has_content != GSF_XML_NO_CONTENT ||
@@ -757,10 +800,20 @@ gsf_xml_in_doc_new (GsfXMLInNode const *nodes, GsfXMLInNS const *ns)
 		} else {
 			node = g_new0 (GsfXMLInNodeInternal, 1);
 			node->pub = *e_node;
+			/* WARNING VILE HACK :
+			 * The api in 1.8.2 passed has_content as a boolean.
+			 * Too many things still use that to change yet.  We
+			 * edit the bool here to be GSF_CONTENT_NONE or
+			 * GSF_XML_CONTENT and try to ignore SHARED_CONTENT */
+			if (node->pub.has_content != 0 &&
+			    node->pub.has_content != GSF_XML_SHARED_CONTENT)
+				node->pub.has_content = GSF_XML_CONTENT;
 			node->groups = NULL;
 			g_hash_table_insert (doc->symbols,
 				(gpointer)node->pub.id, node);
 		}
+		if (e_node == nodes) /* first valid node is the root */
+			doc->root_node = node;
 
 		tmp = g_hash_table_lookup (doc->symbols, node->pub.parent_id);
 		if (tmp != NULL) {
@@ -828,15 +881,14 @@ gsf_xml_in_push_state (GsfXMLIn *xin, GsfXMLInDoc const *doc,
 	g_return_if_fail (doc->root_node != NULL);
 
 	ext = g_new (GsfXMLInExtension, 1);
-	ext->old_doc	= xin->doc;
-	ext->old_state	= xin->user_state;
-	ext->dtor	= dtor;
-	ext->from_unknown = state->from_unknown_handler;
-
-	xin->doc = doc;
-	xin->user_state = new_state;
-
-	push_child (state, &doc->root_node->pub, NULL, attrs, ext);
+	ext->doc	  = doc;
+	ext->state	  = new_state;
+	ext->dtor	  = dtor;
+	if (!(ext->from_unknown = state->from_unknown_handler)) {
+		GsfXMLInNodeInternal *node = (GsfXMLInNodeInternal *) xin->node;
+		node->extensions = g_slist_prepend (node->extensions, ext);
+	} else
+		push_child (state, &doc->root_node->pub, NULL, attrs, ext);
 }
 
 /**
