@@ -82,7 +82,7 @@ od_get_meta_prop (GsfXMLIn *xin, char const *prop_name, GType g_type)
 {
 	GValue *res = g_new0 (GValue, 1);
 	if (gsf_xml_gvalue_from_str (res, g_type, xin->content->str))
-		gsf_doc_meta_data_insert (((GsfOOMetaIn *)xin)->md,
+		gsf_doc_meta_data_insert (((GsfOOMetaIn *)xin->user_state)->md,
 			g_strdup (prop_name), res);
 	else
 		g_free (res);
@@ -126,7 +126,7 @@ doc_stats
 #endif
 
 
-static GsfXMLInNode gsf_opendoc_meta_dtd[] = {
+static GsfXMLInNode const gsf_opendoc_meta_dtd[] = {
   GSF_XML_IN_NODE_FULL (START, START, -1, NULL, FALSE, FALSE, TRUE, NULL, NULL, 0),
   GSF_XML_IN_NODE (START, META, OO_NS_OFFICE, "meta", FALSE, NULL, NULL),
     /* OpenDocument TAGS */
